from abc import ABC, abstractmethod
from typing import Any, Collection


class Specification(ABC):
    @abstractmethod
    def is_satisfied_by(self, obj: Any) -> bool:
        raise NotImplementedError

    @abstractmethod
    def to_collection(self) -> Collection:
        raise NotImplementedError

    def And(self, specification: "Specification") -> "Specification":
        from fractal.core.specifications.generic.collections import AndSpecification

        return AndSpecification([self, specification])

    def Or(self, specification: "Specification") -> "Specification":
        from fractal.core.specifications.generic.collections import OrSpecification

        return OrSpecification([self, specification])

    def __str__(self):
        return self.__class__.__name__

    @staticmethod
    def Not(specification: "Specification") -> "Specification":
        from fractal.core.specifications.generic.operators import NotSpecification

        return NotSpecification(specification)

    @staticmethod
    def parse(**kwargs):
        def __parse(**kwargs):
            for k, v in kwargs.items():
                if "__" not in k:
                    from fractal.core.specifications.generic.operators import (
                        EqualsSpecification,
                    )

                    yield EqualsSpecification(k, v)
                else:
                    field, op = k.split("__")
                    Spec = None
                    if op == "equals":
                        from fractal.core.specifications.generic.operators import (
                            EqualsSpecification as Spec,
                        )
                    elif op == "in":
                        from fractal.core.specifications.generic.operators import (
                            InSpecification as Spec,
                        )
                    elif op == "contains":
                        from fractal.core.specifications.generic.operators import (
                            ContainsSpecification as Spec,
                        )
                    elif op == "lt":
                        from fractal.core.specifications.generic.operators import (
                            LessThenSpecification as Spec,
                        )
                    elif op == "lte":
                        from fractal.core.specifications.generic.operators import (
                            LessThenEqualSpecification as Spec,
                        )
                    elif op == "gt":
                        from fractal.core.specifications.generic.operators import (
                            GreaterThenSpecification as Spec,
                        )
                    elif op == "gte":
                        from fractal.core.specifications.generic.operators import (
                            GreaterThenEqualSpecification as Spec,
                        )
                    if Spec:
                        yield Spec(field, v)

        specs = list(__parse(**kwargs))
        if len(specs) > 1:
            from fractal.core.specifications.generic.collections import AndSpecification

            return AndSpecification(specs)
        elif len(specs) == 1:
            return specs[0]
        return None



from abc import ABC, abstractmethod
from typing import Type

from fractal.core.event_sourcing.event import SendingEvent


class EventProjector(ABC):
    @abstractmethod
    def project(self, id: str, event: Type[SendingEvent]):
        pass



from fastapi import APIRouter

from fractal.contrib.fastapi.routers import Routes
from fractal.contrib.fastapi.routers.domain.models import AdapterInfo, Info
from fractal.core.utils.application_context import ApplicationContext


def inject_default_routes(context: ApplicationContext):
    router = APIRouter()

    @router.get(Routes.ROOT)
    def read_root():
        return {"FastAPI": "Example"}

    @router.get(Routes.INFO, responses={200: {"model": Info}})
    def info():
        adapters = list(context.adapters())
        data = [
            AdapterInfo(
                adapter=adapter.__class__.__name__,
                status=False,
            )
            for adapter in adapters
        ]
        for i in range(len(data)):
            try:
                data[i].status_ok = adapters[i].is_healthy()
            except Exception as e:
                context.logger.error(e)
        return data

    @router.get(Routes.HEALTHZ)
    def healthz():
        return "ok"

    return router



from dataclasses import make_dataclass

from fractal.core.specifications.generic.operators import (
    ContainsSpecification,
    EqualsSpecification,
    GreaterThenEqualSpecification,
    GreaterThenSpecification,
    InSpecification,
    LessThenEqualSpecification,
    LessThenSpecification,
    NotSpecification,
)


def test_not_specification():
    spec = NotSpecification(EqualsSpecification("id", 2))
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_in_specification():
    spec = InSpecification("id", [1, 2, 3])
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_equals_specification():
    spec = EqualsSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_less_then_specification():
    spec = LessThenSpecification("id", 2)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_less_then_equal_specification():
    spec = LessThenEqualSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_greater_then_specification():
    spec = GreaterThenSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=2)))


def test_greater_then_equal_specification():
    spec = GreaterThenEqualSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_contains_specification():
    spec = ContainsSpecification("name", "a")
    DC = make_dataclass("DC", [("name", str)])
    assert spec.is_satisfied_by(DC(**dict(name="fractal")))



import logging

import sentry_sdk
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware
from starlette.requests import Request
from starlette.responses import JSONResponse

from fractal.contrib.fastapi.exceptions.error_message import ErrorMessage
from fractal.core.exceptions import DomainException
from fractal.core.utils.settings import Settings


def install_fastapi(settings: Settings):
    app = FastAPI(root_path=settings.OPENAPI_PREFIX_PATH)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=settings.ALLOW_ORIGINS,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    sentry_sdk.init(dsn=settings.SENTRY_DSN)
    app.add_middleware(
        SentryAsgiMiddleware,
    )

    @app.exception_handler(DomainException)
    def unicorn_exception_handler(request: Request, exc: DomainException):
        logger = logging.getLogger("app")
        if isinstance(exc.status_code, int) and 400 <= exc.status_code < 500:
            logger.warning(f"{exc.code} - {exc.message}")
        else:
            logger.error(f"{exc.code} - {exc.message}")

        return JSONResponse(
            status_code=exc.status_code,
            content=ErrorMessage(
                code=exc.code,
                message=exc.message,
            ).dict(),
            headers=exc.headers,
        )

    return app



from fractal.core.exceptions import DomainException


class TokenInvalidException(DomainException):
    code = "TOKEN_INVALID"
    status_code = 403


class TokenExpiredException(DomainException):
    code = "TOKEN_EXPIRED"
    status_code = 403



from abc import ABC, abstractmethod

from fractal.core.event_sourcing.event_stream import EventStream


class EventStore(ABC):
    @abstractmethod
    def commit(self, event_stream: EventStream, aggregate: str, version: int):
        raise NotImplementedError

    @abstractmethod
    def get_event_stream(self) -> EventStream:
        raise NotImplementedError

    @abstractmethod
    def is_healthy(self) -> bool:
        raise NotImplementedError



from abc import abstractmethod
from functools import wraps
from typing import Set

from fractal.contrib.roles.models import Role
from fractal.contrib.tokens.services import TokenService
from fractal.core.exceptions import DomainException
from fractal.core.services import Service
from fractal.core.utils.application_context import ApplicationContext


class NoPermissionException(DomainException):
    code = "NO_PERMISSION"
    status_code = 403


class RolesService(Service):
    @abstractmethod
    def verify(self, *args, **kwargs):
        raise NotImplementedError

    @staticmethod
    def verify_roles(user_roles: Set[Role], required_roles: Set[Role]):
        if not RolesService.verify_roles_check(user_roles, required_roles):
            raise NoPermissionException(
                f"Got '{user_roles}' while one of '{[role.value for role in required_roles]}' or higher is required."
            )
        return True

    @staticmethod
    def verify_roles_check(user_roles: Set[Role], required_roles: Set[Role]) -> bool:
        return bool(user_roles & required_roles)


class TokenRolesService(RolesService):
    def __init__(self, token_service: TokenService):
        self.token_service = token_service

    def verify(self, required_roles: Set[Role], *args, **kwargs):
        def decorator(func):
            @wraps(func)
            def wrap(token: str, *args, **kwargs):
                payload = self.token_service.verify(token)
                self.verify_roles(set(payload["roles"]), required_roles)
                return func(token, *args, **kwargs)

            return wrap

        return decorator

    @classmethod
    def install(cls, context: ApplicationContext):
        yield cls(*context.get_parameters(["token_service"]))



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['fractal',
 'fractal.contrib',
 'fractal.contrib.django',
 'fractal.contrib.fastapi',
 'fractal.contrib.fastapi.exceptions',
 'fractal.contrib.fastapi.routers',
 'fractal.contrib.fastapi.routers.domain',
 'fractal.contrib.fastapi.utils',
 'fractal.contrib.mongo',
 'fractal.contrib.roles',
 'fractal.contrib.tokens',
 'fractal.core',
 'fractal.core.command_bus',
 'fractal.core.event_sourcing',
 'fractal.core.event_sourcing.projectors',
 'fractal.core.exceptions',
 'fractal.core.models',
 'fractal.core.process',
 'fractal.core.process.actions',
 'fractal.core.repositories',
 'fractal.core.services',
 'fractal.core.specifications',
 'fractal.core.specifications.generic',
 'fractal.core.utils']

package_data = \
{'': ['*']}

install_requires = \
['python-dotenv']

extras_require = \
{'test': ['black', 'flake8', 'isort']}

setup(name='fractal-toolkit',
      version='0.1.8',
      description='Fractal is a scaffolding toolkit for building SOLID logic for your Python applications.',
      author='Douwe van der Meij',
      author_email='douwe@karibu-online.nl',
      url='https://github.com/douwevandermeij/fractal',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.8',
     )



import jwt
from fastapi import Depends
from fastapi.security import OAuth2PasswordBearer
from jwt import DecodeError, ExpiredSignatureError

from fractal.core.exceptions import DomainException


class TokenInvalidException(DomainException):
    code = "TOKEN_INVALID"
    status_code = 401
    headers = {"WWW-Authenticate": "Bearer"}

    def __init__(self, message="The supplied token is invalid!", *args, **kwargs):
        super(TokenInvalidException, self).__init__(message, *args, **kwargs)


class TokenExpiredException(DomainException):
    code = "TOKEN_EXPIRED"
    status_code = 401
    headers = {"WWW-Authenticate": "Bearer"}


oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")


class JwtUtils:
    JWT_PUBLIC_KEY = ""

    def get_token_payload(self, token: str = Depends(oauth2_scheme)):
        try:
            payload = jwt.decode(token, self.JWT_PUBLIC_KEY, algorithms="RS256")
        except DecodeError:
            raise TokenInvalidException()
        except ExpiredSignatureError:
            raise TokenExpiredException("The supplied token is expired!")
        if payload["typ"] not in ["access", "refresh"]:
            raise TokenInvalidException()
        return payload
